@@ -32,25 +32,33 @@ from .ctap import CtapDevice, CtapError
 from .ctap1 import Ctap1, APDU, ApduError
 from .ctap2 import Ctap2, AssertionResponse, Info
 from .ctap2.pin import ClientPin, PinProtocol
-from .ctap2.extensions import Ctap2Extension
+from .ctap2.extensions import (
+    Ctap2Extension,
+    AuthenticationExtensionProcessor,
+)
 from .webauthn import (
     Aaguid,
     AttestationObject,
     CollectedClientData,
+    PublicKeyCredentialRpEntity,
+    PublicKeyCredentialDescriptor,
     PublicKeyCredentialCreationOptions,
     PublicKeyCredentialRequestOptions,
+    AuthenticationExtensionsClientOutputs,
     AuthenticatorSelectionCriteria,
     UserVerificationRequirement,
     AuthenticatorAttestationResponse,
     AuthenticatorAssertionResponse,
     AttestationConveyancePreference,
+    ResidentKeyRequirement,
+    _as_cbor,
 )
 from .cose import ES256
 from .rpid import verify_rp_id
 from .utils import sha256
 from enum import IntEnum, unique
-from urllib.parse import urlparse
 from dataclasses import replace
+from urllib.parse import urlparse
 from threading import Timer, Event
 from typing import (
     Type,
@@ -59,6 +67,8 @@ from typing import (
     Optional,
     Mapping,
     Sequence,
+    Tuple,
+    overload,
 )
 
 import abc
@@ -70,8 +80,12 @@ logger = logging.getLogger(__name__)
 
 
 class ClientError(Exception):
+    """Base error raised by clients."""
+
     @unique
     class ERR(IntEnum):
+        """Error codes for ClientError."""
+
         OTHER_ERROR = 1
         BAD_REQUEST = 2
         CONFIGURATION_UNSUPPORTED = 3
@@ -132,6 +146,8 @@ def _ctap2client_err(e, err_cls=ClientError):
 
 
 class PinRequiredError(ClientError):
+    """Raised when a call cannot be completed without providing PIN."""
+
     def __init__(
         self, code=ClientError.ERR.BAD_REQUEST, cause="PIN required but not provided"
     ):
@@ -185,10 +201,14 @@ class AssertionSelection:
     """
 
     def __init__(
-        self, client_data: CollectedClientData, assertions: Sequence[AssertionResponse]
+        self,
+        client_data: CollectedClientData,
+        assertions: Sequence[AssertionResponse],
+        extension_results=None,
     ):
         self._client_data = client_data
         self._assertions = assertions
+        self._extension_results = extension_results
 
     def get_assertions(self) -> Sequence[AssertionResponse]:
         """Get the raw AssertionResponses available to inspect before selecting one."""
@@ -197,7 +217,7 @@ class AssertionSelection:
     def _get_extension_results(
         self, assertion: AssertionResponse
     ) -> Optional[Mapping[str, Any]]:
-        return None  # Not implemented
+        return self._extension_results
 
     def get_response(self, index: int) -> AuthenticatorAssertionResponse:
         """Get a single response."""
@@ -214,6 +234,8 @@ class AssertionSelection:
 
 
 class WebAuthnClient(abc.ABC):
+    """Base class for a WebAuthn client, supporting registration and authentication."""
+
     @abc.abstractmethod
     def make_credential(
         self,
@@ -292,11 +314,25 @@ class _ClientBackend(abc.ABC):
         raise NotImplementedError()
 
     @abc.abstractmethod
-    def do_make_credential(self, *args) -> AuthenticatorAttestationResponse:
+    def do_make_credential(
+        self,
+        options: PublicKeyCredentialCreationOptions,
+        client_data: CollectedClientData,
+        rp: PublicKeyCredentialRpEntity,
+        rp_id: str,
+        enterprise_rpid_list: Optional[Sequence[str]],
+        event: Event,
+    ) -> AuthenticatorAttestationResponse:
         raise NotImplementedError()
 
     @abc.abstractmethod
-    def do_get_assertion(self, *args) -> AssertionSelection:
+    def do_get_assertion(
+        self,
+        options: PublicKeyCredentialRequestOptions,
+        client_data: CollectedClientData,
+        rp_id: str,
+        event: Event,
+    ) -> AssertionSelection:
         raise NotImplementedError()
 
 
@@ -319,30 +355,32 @@ class _Ctap1ClientBackend(_ClientBackend):
 
     def do_make_credential(
         self,
+        options,
         client_data,
         rp,
-        user,
-        key_params,
-        exclude_list,
-        extensions,
-        rk,
-        user_verification,
-        enterprise_attestation,
+        rp_id,
+        enterprise_rpid_list,
         event,
     ):
+        key_params = options.pub_key_cred_params
+        exclude_list = options.exclude_credentials
+        selection = options.authenticator_selection or AuthenticatorSelectionCriteria()
+        rk = selection.require_resident_key
+        user_verification = selection.user_verification
+
         if (
             rk
             or user_verification == UserVerificationRequirement.REQUIRED
             or ES256.ALGORITHM not in [p.alg for p in key_params]
-            or enterprise_attestation
+            or options.attestation == AttestationConveyancePreference.ENTERPRISE
         ):
             raise CtapError(CtapError.ERR.UNSUPPORTED_OPTION)
 
-        app_param = sha256(rp["id"].encode())
+        app_param = sha256(rp_id.encode())
 
         dummy_param = b"\0" * 32
         for cred in exclude_list or []:
-            key_handle = cred["id"]
+            key_handle = cred.id
             try:
                 self.ctap1.authenticate(dummy_param, app_param, key_handle, True)
                 raise ClientError.ERR.OTHER_ERROR()  # Shouldn't happen
@@ -372,18 +410,19 @@ class _Ctap1ClientBackend(_ClientBackend):
         return AuthenticatorAttestationResponse(
             client_data,
             AttestationObject.create(att_obj.fmt, att_obj.auth_data, att_obj.att_stmt),
-            {},
+            AuthenticationExtensionsClientOutputs({}),
         )
 
     def do_get_assertion(
         self,
+        options,
         client_data,
         rp_id,
-        allow_list,
-        extensions,
-        user_verification,
         event,
     ):
+        allow_list = options.allow_credentials
+        user_verification = options.user_verification
+
         if user_verification == UserVerificationRequirement.REQUIRED or not allow_list:
             raise CtapError(CtapError.ERR.UNSUPPORTED_OPTION)
 
@@ -398,7 +437,7 @@ class _Ctap1ClientBackend(_ClientBackend):
                     self.ctap1.authenticate,
                     client_param,
                     app_param,
-                    cred["id"],
+                    cred.id,
                 )
                 assertions = [AssertionResponse.from_ctap1(app_param, cred, auth_resp)]
                 return AssertionSelection(client_data, assertions)
@@ -413,28 +452,38 @@ class _Ctap2ClientAssertionSelection(AssertionSelection):
         self,
         client_data: CollectedClientData,
         assertions: Sequence[AssertionResponse],
-        extensions: Sequence[Ctap2Extension],
-        pin_token: Optional[str],
-        pin_protocol: Optional[PinProtocol],
+        extensions: Sequence[AuthenticationExtensionProcessor],
+        pin_token: Optional[bytes],
     ):
         super().__init__(client_data, assertions)
         self._extensions = extensions
         self._pin_token = pin_token
-        self._pin_protocol = pin_protocol
 
     def _get_extension_results(self, assertion):
         # Process extenstion outputs
         extension_outputs = {}
         try:
             for ext in self._extensions:
-                output = ext.process_get_output(
-                    assertion, self._pin_token, self._pin_protocol
-                )
-                if output is not None:
+                output = ext.prepare_outputs(assertion, self._pin_token)
+                if output:
                     extension_outputs.update(output)
         except ValueError as e:
             raise ClientError.ERR.CONFIGURATION_UNSUPPORTED(e)
-        return extension_outputs
+        return AuthenticationExtensionsClientOutputs(extension_outputs)
+
+
+@overload
+def _cbor_list(values: Sequence) -> list: ...
+
+
+@overload
+def _cbor_list(values: None) -> None: ...
+
+
+def _cbor_list(values):
+    if not values:
+        return None
+    return [_as_cbor(v) for v in values]
 
 
 class _Ctap2ClientBackend(_ClientBackend):
@@ -442,13 +491,68 @@ class _Ctap2ClientBackend(_ClientBackend):
         self,
         device: CtapDevice,
         user_interaction: UserInteraction,
-        extensions: Sequence[Type[Ctap2Extension]],
+        extension_types: Sequence[Type[Ctap2Extension]],
+        extensions: Sequence[Ctap2Extension],
     ):
         self.ctap2 = Ctap2(device)
         self.info = self.ctap2.info
-        self.extensions = extensions
+        self._extension_types = extension_types
+        self._extensions = extensions
         self.user_interaction = user_interaction
 
+    def _get_extensions(self) -> Sequence[Ctap2Extension]:
+        if self._extensions:
+            return self._extensions
+        return [ext(self.ctap2) for ext in self._extension_types]
+
+    def _filter_creds(
+        self, rp_id, cred_list, pin_protocol, pin_token, event, on_keepalive
+    ):
+        # Filter out credential IDs which are too long
+        max_len = self.info.max_cred_id_length
+        if max_len:
+            cred_list = [c for c in cred_list if len(c.id) <= max_len]
+
+        max_creds = self.info.max_creds_in_list or 1
+        chunks = [
+            cred_list[i : i + max_creds] for i in range(0, len(cred_list), max_creds)
+        ]
+
+        client_data_hash = b"\0" * 32
+        if pin_token:
+            pin_auth = pin_protocol.authenticate(pin_token, client_data_hash)
+            version = pin_protocol.VERSION
+        else:
+            pin_auth = None
+            version = None
+
+        for chunk in chunks:
+            try:
+                assertions = self.ctap2.get_assertions(
+                    rp_id,
+                    client_data_hash,
+                    _cbor_list(chunk),
+                    None,
+                    {"up": False},
+                    pin_auth,
+                    version,
+                    event=event,
+                    on_keepalive=on_keepalive,
+                )
+                if len(chunk) == 1:
+                    # Credential ID might be omitted from assertions
+                    return chunk[0]
+                else:
+                    return PublicKeyCredentialDescriptor(**assertions[0].credential)
+            except CtapError as e:
+                if e.code == CtapError.ERR.NO_CREDENTIALS:
+                    # All creds in chunk are discarded
+                    continue
+                raise
+
+        # No matches found
+        return None
+
     def selection(self, event):
         if "FIDO_2_1" in self.info.versions:
             self.ctap2.selection(event=event)
@@ -472,13 +576,17 @@ class _Ctap2ClientBackend(_ClientBackend):
                     return
                 raise
 
-    def _should_use_uv(self, user_verification, mc):
+    def _should_use_uv(self, user_verification, permissions):
         uv_supported = any(
             k in self.info.options for k in ("uv", "clientPin", "bioEnroll")
         )
         uv_configured = any(
             self.info.options.get(k) for k in ("uv", "clientPin", "bioEnroll")
         )
+        mc = ClientPin.PERMISSION.MAKE_CREDENTIAL & permissions != 0
+        additional_perms = permissions & ~(
+            ClientPin.PERMISSION.MAKE_CREDENTIAL | ClientPin.PERMISSION.GET_ASSERTION
+        )
 
         if (
             user_verification == UserVerificationRequirement.REQUIRED
@@ -495,6 +603,8 @@ class _Ctap2ClientBackend(_ClientBackend):
             return True
         elif mc and uv_configured and not self.info.options.get("makeCredUvNotRqd"):
             return True
+        elif uv_configured and additional_perms:
+            return True
         return False
 
     def _get_token(
@@ -524,110 +634,184 @@ class _Ctap2ClientBackend(_ClientBackend):
         )
 
     def _get_auth_params(
-        self, client_data, rp_id, user_verification, permissions, event, on_keepalive
+        self, pin_protocol, rp_id, user_verification, permissions, event, on_keepalive
     ):
-        mc = client_data.type == CollectedClientData.TYPE.CREATE
         self.info = self.ctap2.get_info()  # Make sure we have "fresh" info
 
-        pin_protocol = None
         pin_token = None
-        pin_auth = None
         internal_uv = False
-        if self._should_use_uv(user_verification, mc) or permissions:
-            client_pin = ClientPin(self.ctap2)
-            allow_internal_uv = not permissions
-            permissions |= (
-                ClientPin.PERMISSION.MAKE_CREDENTIAL
-                if mc
-                else ClientPin.PERMISSION.GET_ASSERTION
+        if self._should_use_uv(user_verification, permissions):
+            client_pin = ClientPin(self.ctap2, pin_protocol)
+            allow_internal_uv = (
+                permissions
+                & ~(
+                    ClientPin.PERMISSION.MAKE_CREDENTIAL
+                    | ClientPin.PERMISSION.GET_ASSERTION
+                )
+                == 0
             )
             pin_token = self._get_token(
                 client_pin, permissions, rp_id, event, on_keepalive, allow_internal_uv
             )
-            if pin_token:
-                pin_protocol = client_pin.protocol
-                pin_auth = client_pin.protocol.authenticate(pin_token, client_data.hash)
-            else:
+            if not pin_token:
                 internal_uv = True
-        return pin_protocol, pin_token, pin_auth, internal_uv
+        return pin_token, internal_uv
 
     def do_make_credential(
         self,
+        options,
         client_data,
         rp,
-        user,
-        key_params,
-        exclude_list,
-        extensions,
-        rk,
-        user_verification,
-        enterprise_attestation,
+        rp_id,
+        enterprise_rpid_list,
         event,
     ):
-        if exclude_list:
-            # Filter out credential IDs which are too long
-            max_len = self.info.max_cred_id_length
-            if max_len:
-                exclude_list = [e for e in exclude_list if len(e) <= max_len]
-
-            # Reject the request if too many credentials remain.
-            max_creds = self.info.max_creds_in_list
-            if max_creds and len(exclude_list) > max_creds:
-                raise ClientError.ERR.BAD_REQUEST("exclude_list too long")
-
-        # Process extensions
-        client_inputs = extensions or {}
-        extension_inputs = {}
-        used_extensions = []
-        permissions = ClientPin.PERMISSION(0)
-        try:
-            for ext in [cls(self.ctap2) for cls in self.extensions]:
-                auth_input, req_perms = ext.process_create_input_with_permissions(
-                    client_inputs
-                )
-                if auth_input is not None:
-                    used_extensions.append(ext)
-                    permissions |= req_perms
-                    extension_inputs[ext.NAME] = auth_input
-        except ValueError as e:
-            raise ClientError.ERR.CONFIGURATION_UNSUPPORTED(e)
+        user = options.user
+        key_params = options.pub_key_cred_params
+        exclude_list = options.exclude_credentials
+        selection = options.authenticator_selection or AuthenticatorSelectionCriteria()
+        user_verification = selection.user_verification
 
         on_keepalive = _user_keepalive(self.user_interaction)
 
-        # Handle auth
-        pin_protocol, pin_token, pin_auth, internal_uv = self._get_auth_params(
-            client_data, rp["id"], user_verification, permissions, event, on_keepalive
-        )
+        # Handle enterprise attestation
+        enterprise_attestation = None
+        if options.attestation == AttestationConveyancePreference.ENTERPRISE:
+            if self.info.options.get("ep"):
+                if enterprise_rpid_list is not None:
+                    # Platform facilitated
+                    if rp_id in enterprise_rpid_list:
+                        enterprise_attestation = 2
+                else:
+                    # Vendor facilitated
+                    enterprise_attestation = 1
 
-        if not (rk or internal_uv):
-            options = None
+        # Negotiate PIN/UV protocol version
+        for proto in ClientPin.PROTOCOLS:
+            if proto.VERSION in self.info.pin_uv_protocols:
+                pin_protocol: Optional[PinProtocol] = proto()
+                break
         else:
-            options = {}
-            if rk:
-                options["rk"] = True
-            if internal_uv:
-                options["uv"] = True
+            pin_protocol = None
 
-        att_obj = self.ctap2.make_credential(
-            client_data.hash,
-            rp,
-            user,
-            key_params,
-            exclude_list or None,
-            extension_inputs or None,
-            options,
-            pin_auth,
-            pin_protocol.VERSION if pin_protocol else None,
-            enterprise_attestation,
-            event=event,
-            on_keepalive=on_keepalive,
-        )
+        # Gather UV permissions
+        permissions = ClientPin.PERMISSION.MAKE_CREDENTIAL
+        if exclude_list:
+            # We need this for filtering the exclude_list
+            permissions |= ClientPin.PERMISSION.GET_ASSERTION
+
+        # Initialize extensions and add extension permissions
+        used_extensions = []
+        for e in self._get_extensions():
+            ext = e.make_credential(self.ctap2, options, pin_protocol)
+            if ext:
+                used_extensions.append(ext)
+                permissions |= ext.permissions
+
+        def _do_make():
+            # Handle auth
+            pin_token, internal_uv = self._get_auth_params(
+                pin_protocol, rp_id, user_verification, permissions, event, on_keepalive
+            )
+
+            if exclude_list:
+                exclude_cred = self._filter_creds(
+                    rp_id, exclude_list, pin_protocol, pin_token, event, on_keepalive
+                )
+                # We know the request will fail if exclude_cred is not None here
+                # BUT DO NOT FAIL EARLY! We still need to prompt for UP, so we keep
+                # processing the request
+            else:
+                exclude_cred = None
+
+            # Process extensions
+            extension_inputs = {}
+            try:
+                for ext in used_extensions:
+                    auth_input = ext.prepare_inputs(pin_token)
+                    if auth_input:
+                        extension_inputs.update(auth_input)
+            except ValueError as e:
+                raise ClientError.ERR.CONFIGURATION_UNSUPPORTED(e)
+
+            can_rk = self.info.options.get("rk")
+            rk = selection.resident_key == ResidentKeyRequirement.REQUIRED or (
+                selection.resident_key == ResidentKeyRequirement.PREFERRED and can_rk
+            )
+
+            if not (rk or internal_uv):
+                options = None
+            else:
+                options = {}
+                if rk:
+                    if not can_rk:
+                        raise ClientError.ERR.CONFIGURATION_UNSUPPORTED(
+                            "Resident key not supported"
+                        )
+                    options["rk"] = True
+                if internal_uv:
+                    options["uv"] = True
+
+            # Calculate pin_auth
+            client_data_hash = client_data.hash
+            if pin_protocol and pin_token:
+                pin_auth: Tuple[Optional[bytes], Optional[int]] = (
+                    pin_protocol.authenticate(pin_token, client_data_hash),
+                    pin_protocol.VERSION,
+                )
+            else:
+                pin_auth = (None, None)
+
+            # Perform make credential
+            return (
+                self.ctap2.make_credential(
+                    client_data_hash,
+                    _as_cbor(replace(rp, id=rp_id)),
+                    _as_cbor(user),
+                    _cbor_list(key_params),
+                    [_as_cbor(exclude_cred)] if exclude_cred else None,
+                    extension_inputs or None,
+                    options,
+                    *pin_auth,
+                    enterprise_attestation,
+                    event=event,
+                    on_keepalive=on_keepalive,
+                ),
+                pin_token,
+            )
+
+        dev = self.ctap2.device
+        reconnected = False
+        while True:
+            try:
+                att_obj, pin_token = _do_make()
+                break
+            except CtapError as e:
+                # The Authenticator may still require UV, try again
+                if (
+                    e.code == CtapError.ERR.PUAT_REQUIRED
+                    and user_verification == UserVerificationRequirement.DISCOURAGED
+                ):
+                    user_verification = UserVerificationRequirement.REQUIRED
+                    continue
+                # NFC may require reconnect
+                connect = getattr(dev, "connect", None)
+                if (
+                    e.code == CtapError.ERR.PIN_AUTH_BLOCKED
+                    and connect
+                    and not reconnected
+                ):
+                    dev.close()
+                    connect()
+                    reconnected = True  # We only want to try this once
+                    continue
+                raise
 
         # Process extenstion outputs
         extension_outputs = {}
         try:
             for ext in used_extensions:
-                output = ext.process_create_output(att_obj, pin_token, pin_protocol)
+                output = ext.prepare_outputs(att_obj, pin_token)
                 if output is not None:
                     extension_outputs.update(output)
         except ValueError as e:
@@ -636,74 +820,125 @@ class _Ctap2ClientBackend(_ClientBackend):
         return AuthenticatorAttestationResponse(
             client_data,
             AttestationObject.create(att_obj.fmt, att_obj.auth_data, att_obj.att_stmt),
-            extension_outputs,
+            AuthenticationExtensionsClientOutputs(extension_outputs),
         )
 
     def do_get_assertion(
         self,
+        options,
         client_data,
         rp_id,
-        allow_list,
-        extensions,
-        user_verification,
         event,
     ):
-        if allow_list:
-            # Filter out credential IDs which are too long
-            max_len = self.info.max_cred_id_length
-            if max_len:
-                allow_list = [e for e in allow_list if len(e) <= max_len]
-            if not allow_list:
-                raise CtapError(CtapError.ERR.NO_CREDENTIALS)
-
-            # Reject the request if too many credentials remain.
-            max_creds = self.info.max_creds_in_list
-            if max_creds and len(allow_list) > max_creds:
-                raise ClientError.ERR.BAD_REQUEST("allow_list too long")
-
-        # Process extensions
-        client_inputs = extensions or {}
-        extension_inputs = {}
-        used_extensions = []
-        permissions = ClientPin.PERMISSION(0)
-        try:
-            for ext in [cls(self.ctap2) for cls in self.extensions]:
-                auth_input, req_perms = ext.process_get_input_with_permissions(
-                    client_inputs
-                )
-                if auth_input is not None:
-                    used_extensions.append(ext)
-                    permissions |= req_perms
-                    extension_inputs[ext.NAME] = auth_input
-        except ValueError as e:
-            raise ClientError.ERR.CONFIGURATION_UNSUPPORTED(e)
+        rp_id = options.rp_id
+        allow_list = options.allow_credentials
+        user_verification = options.user_verification
 
         on_keepalive = _user_keepalive(self.user_interaction)
 
-        pin_protocol, pin_token, pin_auth, internal_uv = self._get_auth_params(
-            client_data, rp_id, user_verification, permissions, event, on_keepalive
-        )
-        options = {"uv": True} if internal_uv else None
+        # Negotiate PIN/UV protocol version
+        for proto in ClientPin.PROTOCOLS:
+            if proto.VERSION in self.info.pin_uv_protocols:
+                pin_protocol: Optional[PinProtocol] = proto()
+                break
+        else:
+            pin_protocol = None
 
-        assertions = self.ctap2.get_assertions(
-            rp_id,
-            client_data.hash,
-            allow_list or None,
-            extension_inputs or None,
-            options,
-            pin_auth,
-            pin_protocol.VERSION if pin_protocol else None,
-            event=event,
-            on_keepalive=on_keepalive,
-        )
+        # Gather UV permissions
+        permissions = ClientPin.PERMISSION.GET_ASSERTION
 
-        return _Ctap2ClientAssertionSelection(
-            client_data,
-            assertions,
-            used_extensions,
-            pin_token,
-            pin_protocol,
-        )
+        # Initialize extensions and add extension permissions
+        used_extensions = []
+        for e in self._get_extensions():
+            ext = e.get_assertion(self.ctap2, options, pin_protocol)
+            if ext:
+                used_extensions.append(ext)
+                permissions |= ext.permissions
+
+        def _do_auth():
+            # Handle auth
+            pin_token, internal_uv = self._get_auth_params(
+                pin_protocol, rp_id, user_verification, permissions, event, on_keepalive
+            )
+
+            if allow_list:
+                selected_cred = self._filter_creds(
+                    rp_id, allow_list, pin_protocol, pin_token, event, on_keepalive
+                )
+                # We know the request will fail if selected_cred is None here
+                # BUT DO NOT FAIL EARLY! We still need to prompt for UP, so we keep
+                # processing the request
+            else:
+                selected_cred = None
+
+            # Process extensions
+            extension_inputs = {}
+            try:
+                for ext in used_extensions:
+                    inputs = ext.prepare_inputs(selected_cred, pin_token)
+                    if inputs:
+                        extension_inputs.update(inputs)
+            except ValueError as e:
+                raise ClientError.ERR.CONFIGURATION_UNSUPPORTED(e)
+
+            options = {"uv": True} if internal_uv else None
+
+            # Calculate pin_auth
+            client_data_hash = client_data.hash
+            if pin_protocol and pin_token:
+                pin_auth: Tuple[Optional[bytes], Optional[int]] = (
+                    pin_protocol.authenticate(pin_token, client_data_hash),
+                    pin_protocol.VERSION,
+                )
+            else:
+                pin_auth = (None, None)
+
+            if allow_list and not selected_cred:
+                # We still need to send a dummy value if there was an allow_list
+                # but no matches were found:
+                selected_cred = PublicKeyCredentialDescriptor(allow_list[0].type, b"\0")
+
+            # Perform get assertion
+            assertions = self.ctap2.get_assertions(
+                rp_id,
+                client_data_hash,
+                [_as_cbor(selected_cred)] if selected_cred else None,
+                extension_inputs or None,
+                options,
+                *pin_auth,
+                event=event,
+                on_keepalive=on_keepalive,
+            )
+
+            return _Ctap2ClientAssertionSelection(
+                client_data, assertions, used_extensions, pin_token
+            )
+
+        dev = self.ctap2.device
+        reconnected = False
+        while True:
+            try:
+                return _do_auth()
+            except CtapError as e:
+                # The Authenticator may still require UV, try again
+                if (
+                    e.code == CtapError.ERR.PUAT_REQUIRED
+                    and user_verification == UserVerificationRequirement.DISCOURAGED
+                ):
+                    user_verification = UserVerificationRequirement.REQUIRED
+                    continue
+                # NFC may require reconnect
+                connect = getattr(dev, "connect", None)
+                if (
+                    e.code == CtapError.ERR.PIN_AUTH_BLOCKED
+                    and connect
+                    and not reconnected
+                ):
+                    dev.close()
+                    connect()
+                    reconnected = True  # We only want to try this once
+                    continue
+                raise
 
 
 class Fido2Client(WebAuthnClient, _BaseClient):
@@ -722,8 +957,10 @@ class Fido2Client(WebAuthnClient, _BaseClient):
         device: CtapDevice,
         origin: str,
         verify: Callable[[str, str], bool] = verify_rp_id,
+        # TODO 2.0: Replace extension_types with extensions
         extension_types: Sequence[Type[Ctap2Extension]] = _default_extensions(),
         user_interaction: UserInteraction = UserInteraction(),
+        extensions: Sequence[Ctap2Extension] = [],
     ):
         super().__init__(origin, verify)
 
@@ -732,7 +969,7 @@ class Fido2Client(WebAuthnClient, _BaseClient):
 
         try:
             self._backend: _ClientBackend = _Ctap2ClientBackend(
-                device, user_interaction, extension_types
+                device, user_interaction, extension_types, extensions
             )
         except (ValueError, CtapError):
             self._backend = _Ctap1ClientBackend(device, user_interaction)
@@ -747,6 +984,17 @@ class Fido2Client(WebAuthnClient, _BaseClient):
         except CtapError as e:
             raise _ctap2client_err(e)
 
+    def _get_rp_id(self, rp_id: Optional[str]) -> str:
+        if rp_id is None:
+            url = urlparse(self.origin)
+            if url.scheme != "https" or not url.netloc:
+                raise ClientError.ERR.BAD_REQUEST(
+                    "RP ID required for non-https origin."
+                )
+            return url.netloc
+        else:
+            return rp_id
+
     def make_credential(
         self,
         options: PublicKeyCredentialCreationOptions,
@@ -766,44 +1014,22 @@ class Fido2Client(WebAuthnClient, _BaseClient):
             timer.start()
 
         rp = options.rp
-        if rp.id is None:
-            url = urlparse(self.origin)
-            if url.scheme != "https" or not url.netloc:
-                raise ClientError.ERR.BAD_REQUEST(
-                    "RP ID required for non-https origin."
-                )
-            rp = replace(rp, id=url.netloc)
+        rp_id = self._get_rp_id(rp.id)
 
-        logger.debug(f"Register a new credential for RP ID: {rp.id}")
-        self._verify_rp_id(rp.id)
+        logger.debug(f"Register a new credential for RP ID: {rp_id}")
+        self._verify_rp_id(rp_id)
 
         client_data = self._build_client_data(
             CollectedClientData.TYPE.CREATE, options.challenge
         )
 
-        selection = options.authenticator_selection or AuthenticatorSelectionCriteria()
-        enterprise_attestation = None
-        if options.attestation == AttestationConveyancePreference.ENTERPRISE:
-            if self.info.options.get("ep"):
-                if self._enterprise_rpid_list is not None:
-                    # Platform facilitated
-                    if rp.id in self._enterprise_rpid_list:
-                        enterprise_attestation = 2
-                else:
-                    # Vendor facilitated
-                    enterprise_attestation = 1
-
         try:
             return self._backend.do_make_credential(
+                options,
                 client_data,
                 rp,
-                options.user,
-                options.pub_key_cred_params,
-                options.exclude_credentials,
-                options.extensions,
-                selection.require_resident_key,
-                selection.user_verification,
-                enterprise_attestation,
+                rp_id,
+                self._enterprise_rpid_list,
                 event,
             )
         except CtapError as e:
@@ -830,8 +1056,9 @@ class Fido2Client(WebAuthnClient, _BaseClient):
             timer.daemon = True
             timer.start()
 
-        logger.debug(f"Assert a credential for RP ID: {options.rp_id}")
-        self._verify_rp_id(options.rp_id)
+        rp_id = self._get_rp_id(options.rp_id)
+        logger.debug(f"Assert a credential for RP ID: {rp_id}")
+        self._verify_rp_id(rp_id)
 
         client_data = self._build_client_data(
             CollectedClientData.TYPE.GET, options.challenge
@@ -839,11 +1066,9 @@ class Fido2Client(WebAuthnClient, _BaseClient):
 
         try:
             return self._backend.do_get_assertion(
+                options,
                 client_data,
-                options.rp_id,
-                options.allow_credentials,
-                options.extensions,
-                options.user_verification,
+                rp_id,
                 event,
             )
         except CtapError as e:
@@ -859,7 +1084,8 @@ if platform.system().lower() == "windows":
             WinAPI,
             WebAuthNAuthenticatorAttachment,
             WebAuthNUserVerificationRequirement,
-            WebAuthNAttestationConvoyancePreference,
+            WebAuthNAttestationConveyancePreference,
+            WebAuthNEnterpriseAttestation,
         )
     except Exception:  # nosec # TODO: Make this less generic
         pass
@@ -885,18 +1111,24 @@ class WindowsClient(WebAuthnClient, _BaseClient):
         origin: str,
         verify: Callable[[str, str], bool] = verify_rp_id,
         handle=None,
+        allow_hmac_secret=False,
     ):
         super().__init__(origin, verify)
-        self.api = WinAPI(handle)
+        self.api = WinAPI(
+            handle, return_extensions=True, allow_hmac_secret=allow_hmac_secret
+        )
         self.info = Info(
             versions=["U2F_V2", "FIDO_2_0"], extensions=[], aaguid=Aaguid.NONE
         )
 
+        # TODO: Decide how to configure this list.
+        self._enterprise_rpid_list: Optional[Sequence[str]] = None
+
     @staticmethod
     def is_available() -> bool:
         return platform.system().lower() == "windows" and WinAPI.version > 0
 
-    def make_credential(self, options, **kwargs):
+    def make_credential(self, options, event=None):
         """Create a credential using Windows WebAuthN APIs.
 
         :param options: PublicKeyCredentialCreationOptions data.
@@ -914,36 +1146,55 @@ class WindowsClient(WebAuthnClient, _BaseClient):
 
         selection = options.authenticator_selection or AuthenticatorSelectionCriteria()
 
+        enterprise_attestation = WebAuthNEnterpriseAttestation.NONE
+        if options.attestation == AttestationConveyancePreference.ENTERPRISE:
+            attestation = WebAuthNAttestationConveyancePreference.ANY
+            if self.info.options.get("ep"):
+                if self._enterprise_rpid_list is not None:
+                    # Platform facilitated
+                    if options.rp.id in self._enterprise_rpid_list:
+                        enterprise_attestation = (
+                            WebAuthNEnterpriseAttestation.PLATFORM_MANAGED
+                        )
+                else:
+                    # Vendor facilitated
+                    enterprise_attestation = (
+                        WebAuthNEnterpriseAttestation.VENDOR_FACILITATED
+                    )
+        else:
+            attestation = WebAuthNAttestationConveyancePreference.from_string(
+                options.attestation or "none"
+            )
+
         try:
-            result = self.api.make_credential(
+            att_obj, extensions = self.api.make_credential(
                 options.rp,
                 options.user,
                 options.pub_key_cred_params,
                 client_data,
                 options.timeout or 0,
-                selection.require_resident_key or False,
+                selection.resident_key,
                 WebAuthNAuthenticatorAttachment.from_string(
                     selection.authenticator_attachment or "any"
                 ),
                 WebAuthNUserVerificationRequirement.from_string(
                     selection.user_verification or "discouraged"
                 ),
-                WebAuthNAttestationConvoyancePreference.from_string(
-                    options.attestation or "none"
-                ),
+                attestation,
                 options.exclude_credentials,
                 options.extensions,
-                kwargs.get("event"),
+                event,
+                enterprise_attestation,
             )
         except OSError as e:
             raise ClientError.ERR.OTHER_ERROR(e)
 
         logger.info("New credential registered")
         return AuthenticatorAttestationResponse(
-            client_data, AttestationObject(result), {}
+            client_data, att_obj, AuthenticationExtensionsClientOutputs(extensions)
         )
 
-    def get_assertion(self, options, **kwargs):
+    def get_assertion(self, options, event=None):
         """Get assertion using Windows WebAuthN APIs.
 
         :param options: PublicKeyCredentialRequestOptions data.
@@ -960,17 +1211,19 @@ class WindowsClient(WebAuthnClient, _BaseClient):
         )
 
         try:
-            (credential, auth_data, signature, user_id) = self.api.get_assertion(
-                options.rp_id,
-                client_data,
-                options.timeout or 0,
-                WebAuthNAuthenticatorAttachment.ANY,
-                WebAuthNUserVerificationRequirement.from_string(
-                    options.user_verification or "discouraged"
-                ),
-                options.allow_credentials,
-                options.extensions,
-                kwargs.get("event"),
+            (credential, auth_data, signature, user_id, extensions) = (
+                self.api.get_assertion(
+                    options.rp_id,
+                    client_data,
+                    options.timeout or 0,
+                    WebAuthNAuthenticatorAttachment.ANY,
+                    WebAuthNUserVerificationRequirement.from_string(
+                        options.user_verification or "discouraged"
+                    ),
+                    options.allow_credentials,
+                    options.extensions,
+                    event,
+                )
             )
         except OSError as e:
             raise ClientError.ERR.OTHER_ERROR(e)
@@ -986,4 +1239,5 @@ class WindowsClient(WebAuthnClient, _BaseClient):
                     user=user,
                 )
             ],
+            AuthenticationExtensionsClientOutputs(extensions),
         )
